@@ -1,5 +1,6 @@
 #[cfg(feature = "blocking")]
 use block_on_proc::block_on;
+#[cfg(feature = "tags")]
 use minidom::Element;
 use serde_xml_rs as serde_xml;
 use std::collections::HashMap;
@@ -24,9 +25,9 @@ use crate::surf_request::SurfRequest as RequestImpl;
 // use tokio::fs::File;
 
 #[cfg(feature = "with-async-std")]
-use futures::io::AsyncRead;
+use futures_io::{AsyncRead, AsyncWrite};
 #[cfg(feature = "with-tokio")]
-use tokio::io::AsyncRead;
+use tokio::io::{AsyncRead, AsyncWrite};
 
 #[cfg(feature = "sync")]
 use crate::blocking::AttoRequest as RequestImpl;
@@ -49,6 +50,8 @@ use http::HeaderMap;
 
 pub const CHUNK_SIZE: usize = 8_388_608; // 8 Mebibytes, min is 5 (5_242_880);
 
+const DEFAULT_REQUEST_TIMEOUT: Option<Duration> = Some(Duration::from_secs(60));
+
 #[derive(Debug, PartialEq)]
 pub struct Tag {
     key: String,
@@ -88,6 +91,7 @@ pub struct Bucket {
     pub extra_query: Query,
     pub request_timeout: Option<Duration>,
     path_style: bool,
+    listobjects_v2: bool,
 }
 
 fn validate_expiry(expiry_secs: u32) -> Result<()> {
@@ -358,8 +362,9 @@ impl Bucket {
             credentials,
             extra_headers: HeaderMap::new(),
             extra_query: HashMap::new(),
-            request_timeout: None,
+            request_timeout: DEFAULT_REQUEST_TIMEOUT,
             path_style: false,
+            listobjects_v2: true,
         })
     }
 
@@ -382,8 +387,9 @@ impl Bucket {
             credentials: Credentials::anonymous()?,
             extra_headers: HeaderMap::new(),
             extra_query: HashMap::new(),
-            request_timeout: None,
+            request_timeout: DEFAULT_REQUEST_TIMEOUT,
             path_style: false,
+            listobjects_v2: true,
         })
     }
 
@@ -411,8 +417,9 @@ impl Bucket {
             credentials,
             extra_headers: HeaderMap::new(),
             extra_query: HashMap::new(),
-            request_timeout: None,
+            request_timeout: DEFAULT_REQUEST_TIMEOUT,
             path_style: true,
+            listobjects_v2: true,
         })
     }
 
@@ -435,8 +442,9 @@ impl Bucket {
             credentials: Credentials::anonymous()?,
             extra_headers: HeaderMap::new(),
             extra_query: HashMap::new(),
-            request_timeout: None,
+            request_timeout: DEFAULT_REQUEST_TIMEOUT,
             path_style: true,
+            listobjects_v2: true,
         })
     }
 
@@ -633,24 +641,27 @@ impl Bucket {
     /// let credentials = Credentials::default()?;
     /// let bucket = Bucket::new(bucket_name, region, credentials)?;
     /// let mut output_file = File::create("output_file").expect("Unable to create file");
+    /// let mut async_output_file = tokio::fs::File::create("async_output_file").await.expect("Unable to create file");
+    /// #[cfg(feature = "with-async-std")]
+    /// let mut async_output_file = async_std::fs::File::create("async_output_file").await.expect("Unable to create file");
     ///
     /// // Async variant with `tokio` or `async-std` features
-    /// let status_code = bucket.get_object_stream("/test.file", &mut output_file).await?;
+    /// let status_code = bucket.get_object_stream("/test.file", &mut async_output_file).await?;
     ///
     /// // `sync` feature will produce an identical method
     /// #[cfg(feature = "sync")]
     /// let status_code = bucket.get_object_stream("/test.file", &mut output_file)?;
     ///
     /// // Blocking variant, generated with `blocking` feature in combination
-    /// // with `tokio` or `async-std` features.
+    /// // with `tokio` or `async-std` features. Based of the async branch
     /// #[cfg(feature = "blocking")]
-    /// let status_code = bucket.get_object_stream_blocking("/test.file", &mut output_file)?;
+    /// let status_code = bucket.get_object_stream_blocking("/test.file", &mut async_output_file)?;
     /// #
     /// # Ok(())
     /// # }
     /// ```
-    #[maybe_async::maybe_async]
-    pub async fn get_object_stream<T: std::io::Write + Send, S: AsRef<str>>(
+    #[maybe_async::async_impl]
+    pub async fn get_object_stream<T: AsyncWrite + Send + Unpin, S: AsRef<str>>(
         &self,
         path: S,
         writer: &mut T,
@@ -660,6 +671,17 @@ impl Bucket {
         request.response_data_to_writer(writer).await
     }
 
+    #[maybe_async::sync_impl]
+    pub fn get_object_stream<T: std::io::Write + Send, S: AsRef<str>>(
+        &self,
+        path: S,
+        writer: &mut T,
+    ) -> Result<u16> {
+        let command = Command::GetObject;
+        let request = RequestImpl::new(self, path.as_ref(), command);
+        request.response_data_to_writer(writer)
+    }
+
     /// Stream file from local path to s3, generic over T: Write.
     ///
     /// # Example:
@@ -689,7 +711,7 @@ impl Bucket {
     /// #[cfg(feature = "with-async-std")]
     /// let mut path = async_std::fs::File::open(path).await?;
     /// // Async variant with `tokio` or `async-std` features
-    /// // Generic over futures::io::AsyncRead|tokio::io::AsyncRead + Unpin
+    /// // Generic over futures_io::AsyncRead|tokio::io::AsyncRead + Unpin
     /// let status_code = bucket.put_object_stream(&mut path, "/path").await?;
     ///
     /// // `sync` feature will produce an identical method
@@ -729,9 +751,18 @@ impl Bucket {
         reader: &mut R,
         s3_path: &str,
     ) -> Result<u16> {
+        // If the file is smaller CHUNK_SIZE, just do a regular upload.
+        // Otherwise perform a multi-part upload.
+        let first_chunk = crate::utils::read_chunk(reader).await?;
+        if first_chunk.len() < CHUNK_SIZE {
+            let (_, code) = self.put_object(s3_path, first_chunk.as_slice()).await?;
+            return Ok(code);
+        }
+
         let command = Command::InitiateMultipartUpload;
         let request = RequestImpl::new(self, s3_path, command);
         let (data, code) = request.response_data(false).await?;
+
         let msg: InitiateMultipartUploadResponse =
             serde_xml::from_str(std::str::from_utf8(data.as_slice())?)?;
         let path = msg.key;
@@ -739,60 +770,48 @@ impl Bucket {
 
         let mut part_number: u32 = 0;
         let mut etags = Vec::new();
+
+        let mut chunk = first_chunk;
         loop {
-            let chunk = crate::utils::read_chunk(reader).await?;
+            // Upload this chunk
+            part_number += 1;
+            let command = Command::PutObject {
+                // part_number,
+                content: &chunk,
+                content_type: "application/octet-stream",
+                multipart: Some(Multipart::new(part_number, upload_id)), // upload_id: &msg.upload_id,
+            };
+            let request = RequestImpl::new(self, &path, command);
+            let (data, _code) = request.response_data(true).await?;
+            let etag = std::str::from_utf8(data.as_slice())?;
+            etags.push(etag.to_string());
 
             if chunk.len() < CHUNK_SIZE {
-                if part_number == 0 {
-                    // Files is not big enough for multipart upload, going with regular put_object
-                    self.abort_upload(&path, upload_id).await?;
-
-                    self.put_object(s3_path, chunk.as_slice()).await?;
-                } else {
-                    part_number += 1;
-                    let command = Command::PutObject {
-                        // part_number,
-                        content: &chunk,
-                        content_type: "application/octet-stream",
-                        multipart: Some(Multipart::new(part_number, upload_id)), // upload_id: &msg.upload_id,
-                    };
-                    let request = RequestImpl::new(self, &path, command);
-                    let (data, _code) = request.response_data(true).await?;
-                    let etag = std::str::from_utf8(data.as_slice())?;
-                    etags.push(etag.to_string());
-                    let inner_data = etags
-                        .clone()
-                        .into_iter()
-                        .enumerate()
-                        .map(|(i, x)| Part {
-                            etag: x,
-                            part_number: i as u32 + 1,
-                        })
-                        .collect::<Vec<Part>>();
-                    let data = CompleteMultipartUploadData { parts: inner_data };
-                    let complete = Command::CompleteMultipartUpload {
-                        upload_id: &msg.upload_id,
-                        data,
-                    };
-                    let complete_request = RequestImpl::new(self, &path, complete);
-                    let (_data, _code) = complete_request.response_data(false).await?;
-                    // let response = std::str::from_utf8(data.as_slice())?;
-                }
                 break;
-            } else {
-                part_number += 1;
-                let command = Command::PutObject {
-                    // part_number,
-                    content: &chunk,
-                    content_type: "application/octet-stream",
-                    multipart: Some(Multipart::new(part_number, upload_id)), // upload_id: &msg.upload_id,
-                };
-                let request = RequestImpl::new(self, &path, command);
-                let (data, _code) = request.response_data(true).await?;
-                let etag = std::str::from_utf8(data.as_slice())?;
-                etags.push(etag.to_string());
             }
+
+            chunk = crate::utils::read_chunk(reader).await?
         }
+
+        // Finish the upload
+        let inner_data = etags
+            .clone()
+            .into_iter()
+            .enumerate()
+            .map(|(i, x)| Part {
+                etag: x,
+                part_number: i as u32 + 1,
+            })
+            .collect::<Vec<Part>>();
+        let data = CompleteMultipartUploadData { parts: inner_data };
+        let complete = Command::CompleteMultipartUpload {
+            upload_id: &msg.upload_id,
+            data,
+        };
+        let complete_request = RequestImpl::new(self, &path, complete);
+        let (_data, _code) = complete_request.response_data(false).await?;
+        // let response = std::str::from_utf8(data.as_slice())?;
+
         Ok(code)
     }
 
@@ -1231,6 +1250,7 @@ impl Bucket {
     /// # Ok(())
     /// # }
     /// ```
+    #[cfg(feature = "tags")]
     #[maybe_async::maybe_async]
     pub async fn get_object_tagging<S: AsRef<str>>(&self, path: S) -> Result<(Vec<Tag>, u16)> {
         let command = Command::GetObjectTagging {};
@@ -1277,12 +1297,24 @@ impl Bucket {
         start_after: Option<String>,
         max_keys: Option<usize>,
     ) -> Result<(ListBucketResult, u16)> {
-        let command = Command::ListBucket {
-            prefix,
-            delimiter,
-            continuation_token,
-            start_after,
-            max_keys,
+        let command = if self.listobjects_v2 {
+            Command::ListObjectsV2 {
+                prefix,
+                delimiter,
+                continuation_token,
+                start_after,
+                max_keys,
+            }
+        } else {
+            // In the v1 ListObjects request, there is only one "marker"
+            // field that serves as both the initial starting position,
+            // and as the continuation token.
+            Command::ListObjects {
+                prefix,
+                delimiter,
+                marker: std::cmp::max(continuation_token, start_after),
+                max_keys,
+            }
         };
         let request = RequestImpl::new(self, "/", command);
         let (response, status_code) = request.response_data(false).await?;
@@ -1515,7 +1547,8 @@ impl Bucket {
     }
 
     /// Configure bucket to apply this request timeout to all HTTP
-    /// requests, or no (infinity) timeout if `None`.
+    /// requests, or no (infinity) timeout if `None`.  Defaults to
+    /// 30 seconds.
     ///
     /// Only the attohttpc and the Reqwest backends obey this option;
     /// async code may instead await with a timeout.
@@ -1523,6 +1556,20 @@ impl Bucket {
         self.request_timeout = timeout;
     }
 
+    /// Configure bucket to use the older ListObjects API
+    ///
+    /// If your provider doesn't support the ListObjectsV2 interface, set this to
+    /// use the v1 ListObjects interface instead. This is currently needed at least
+    /// for Google Cloud Storage.
+    pub fn set_listobjects_v1(&mut self) {
+        self.listobjects_v2 = false;
+    }
+
+    /// Configure bucket to use the newer ListObjectsV2 API
+    pub fn set_listobjects_v2(&mut self) {
+        self.listobjects_v2 = true;
+    }
+
     /// Get a reference to the name of the S3 bucket.
     pub fn name(&self) -> String {
         self.name.to_string()
@@ -1742,7 +1789,7 @@ mod test {
     }
 
     fn test_gc_bucket() -> Bucket {
-        Bucket::new(
+        let mut bucket = Bucket::new(
             "rust-s3",
             Region::Custom {
                 region: "us-east1".to_owned(),
@@ -1750,7 +1797,9 @@ mod test {
             },
             test_gc_credentials(),
         )
-        .unwrap()
+        .unwrap();
+        bucket.set_listobjects_v1();
+        return bucket;
     }
 
     fn test_minio_bucket() -> Bucket {
@@ -1806,6 +1855,7 @@ mod test {
     }
 
     #[ignore]
+    #[cfg(feature = "tags")]
     #[maybe_async::test(
         feature = "sync",
         async(all(not(feature = "sync"), feature = "with-tokio"), tokio::test),
@@ -1845,10 +1895,18 @@ mod test {
         let (_data, _code) = bucket.delete_object("tagging_test").await.unwrap();
     }
 
+    /// Test multi-part upload
     #[ignore]
     #[maybe_async::test(
         feature = "sync",
-        async(all(not(feature = "sync"), feature = "with-tokio"), tokio::test),
+        async(
+            all(
+                not(feature = "sync"),
+                not(feature = "tokio-rustls-tls"),
+                feature = "with-tokio"
+            ),
+            tokio::test
+        ),
         async(
             all(not(feature = "sync"), feature = "with-async-std"),
             async_std::test
@@ -1860,10 +1918,10 @@ mod test {
         let local_path = "+stream_test_big";
         std::fs::remove_file(remote_path).unwrap_or_else(|_| {});
         let bucket = test_aws_bucket();
-        let test: Vec<u8> = object(10_000_000);
+        let content: Vec<u8> = object(10_000_000);
 
         let mut file = File::create(local_path).unwrap();
-        file.write_all(&test).unwrap();
+        file.write_all(&content).unwrap();
         cfg_if! {
             if #[cfg(feature = "with-tokio")] {
                 let mut reader = tokio::fs::File::open(local_path).await.unwrap();
@@ -1881,14 +1939,13 @@ mod test {
         assert_eq!(code, 200);
         let mut writer = Vec::new();
         let code = bucket
-            .get_object_stream(local_path, &mut writer)
+            .get_object_stream(remote_path, &mut writer)
             .await
             .unwrap();
         assert_eq!(code, 200);
-        assert_eq!(test, writer);
-        let (_body, _code) = bucket.get_object_torrent(remote_path).await.unwrap();
-        // let dummy: Vec<u8> = Vec::new();
-        // Getting a 405 here for some reason
+        assert_eq!(content, writer);
+
+        // let (body, code) = bucket.get_object_torrent(remote_path).await.unwrap();
         // assert_eq!(code, 200);
         // assert_eq!(
         //     body,
@@ -1958,19 +2015,70 @@ mod test {
         std::fs::remove_file(local_path).unwrap_or_else(|_| {});
     }
 
+    /// Test streaming upload, with an object that's smaller than CHUNK_SIZE.
+    /// This should be optimized into a plain PUT, not a multi-part upload.
+    #[ignore]
+    #[maybe_async::test(
+        feature = "sync",
+        async(
+            all(
+                not(feature = "sync"),
+                not(feature = "tokio-rustls-tls"),
+                feature = "with-tokio"
+            ),
+            tokio::test
+        ),
+        async(
+            all(not(feature = "sync"), feature = "with-async-std"),
+            async_std::test
+        )
+    )]
+    async fn streaming_test_put_get_delete_small_object() {
+        init();
+        let remote_path = "+stream_test_small";
+        let bucket = test_gc_bucket();
+        let content: Vec<u8> = object(1000);
+        let mut reader = std::io::Cursor::new(&content);
+
+        #[cfg(feature = "with-async-std")]
+        let mut reader = async_std::io::Cursor::new(&content);
+
+        let code = bucket
+            .put_object_stream(&mut reader, remote_path)
+            .await
+            .unwrap();
+        assert_eq!(code, 200);
+        let mut writer = Vec::new();
+        let code = bucket
+            .get_object_stream(remote_path, &mut writer)
+            .await
+            .unwrap();
+        assert_eq!(code, 200);
+        assert_eq!(content, writer);
+
+        let (_, code) = bucket.delete_object(remote_path).await.unwrap();
+        assert_eq!(code, 204);
+    }
+
     #[cfg(feature = "blocking")]
-    fn put_head_get_delete_object_blocking(bucket: Bucket) {
+    fn put_head_get_list_delete_object_blocking(bucket: Bucket) {
         let s3_path = "/test_blocking.file";
+        let s3_path_2 = "/test_blocking.file2";
+        let s3_path_3 = "/test_blocking.file3";
         let test: Vec<u8> = object(3072);
 
+        // Test PutObject
         let (_data, code) = bucket.put_object_blocking(s3_path, &test).unwrap();
         // println!("{}", std::str::from_utf8(&data).unwrap());
         assert_eq!(code, 200);
+
+        // Test GetObject
         let (data, code) = bucket.get_object_blocking(s3_path).unwrap();
         assert_eq!(code, 200);
         // println!("{}", std::str::from_utf8(&data).unwrap());
         assert_eq!(test, data);
 
+        // Test GetObject with a range
         let (data, code) = bucket
             .get_object_range_blocking(s3_path, 100, Some(1000))
             .unwrap();
@@ -1978,6 +2086,7 @@ mod test {
         // println!("{}", std::str::from_utf8(&data).unwrap());
         assert_eq!(test[100..1001].to_vec(), data);
 
+        // Test HeadObject
         let (head_object_result, code) = bucket.head_object_blocking(s3_path).unwrap();
         assert_eq!(code, 200);
         assert_eq!(
@@ -1985,8 +2094,51 @@ mod test {
             "application/octet-stream".to_owned()
         );
         // println!("{:?}", head_object_result);
+
+        // Put some additional objects, so that we can test ListObjects
+        let (_data, code) = bucket.put_object_blocking(s3_path_2, &test).unwrap();
+        assert_eq!(code, 200);
+        let (_data, code) = bucket.put_object_blocking(s3_path_3, &test).unwrap();
+        assert_eq!(code, 200);
+
+        // Test ListObjects, with continuation
+        let (result, code) = bucket
+            .list_page_blocking(
+                "test_blocking.".to_string(),
+                Some("/".to_string()),
+                None,
+                None,
+                Some(2),
+            )
+            .unwrap();
+        assert_eq!(code, 200);
+        assert_eq!(result.contents.len(), 2);
+        assert_eq!(result.contents[0].key, s3_path[1..]);
+        assert_eq!(result.contents[1].key, s3_path_2[1..]);
+
+        let cont_token = result.next_continuation_token.unwrap();
+
+        let (result, code) = bucket
+            .list_page_blocking(
+                "test_blocking.".to_string(),
+                Some("/".to_string()),
+                Some(cont_token),
+                None,
+                Some(2),
+            )
+            .unwrap();
+        assert_eq!(code, 200);
+        assert_eq!(result.contents.len(), 1);
+        assert_eq!(result.contents[0].key, s3_path_3[1..]);
+        assert!(result.next_continuation_token.is_none());
+
+        // cleanup (and test Delete)
         let (_, code) = bucket.delete_object_blocking(s3_path).unwrap();
         assert_eq!(code, 204);
+        let (_, code) = bucket.delete_object_blocking(s3_path_2).unwrap();
+        assert_eq!(code, 204);
+        let (_, code) = bucket.delete_object_blocking(s3_path_3).unwrap();
+        assert_eq!(code, 204);
     }
 
     #[ignore]
@@ -1996,7 +2148,7 @@ mod test {
     ))]
     #[test]
     fn aws_put_head_get_delete_object_blocking() {
-        put_head_get_delete_object_blocking(test_aws_bucket())
+        put_head_get_list_delete_object_blocking(test_aws_bucket())
     }
 
     #[ignore]
@@ -2006,7 +2158,7 @@ mod test {
     ))]
     #[test]
     fn gc_put_head_get_delete_object_blocking() {
-        put_head_get_delete_object_blocking(test_gc_bucket())
+        put_head_get_list_delete_object_blocking(test_gc_bucket())
     }
 
     #[ignore]
@@ -2016,7 +2168,7 @@ mod test {
     ))]
     #[test]
     fn wasabi_put_head_get_delete_object_blocking() {
-        put_head_get_delete_object_blocking(test_wasabi_bucket())
+        put_head_get_list_delete_object_blocking(test_wasabi_bucket())
     }
 
     #[ignore]
@@ -2036,7 +2188,7 @@ mod test {
             BucketConfiguration::default(),
         )
         .unwrap();
-        put_head_get_delete_object_blocking(test_minio_bucket())
+        put_head_get_list_delete_object_blocking(test_minio_bucket())
     }
 
     #[ignore]
@@ -2046,7 +2198,7 @@ mod test {
     ))]
     #[test]
     fn digital_ocean_put_head_get_delete_object_blocking() {
-        put_head_get_delete_object_blocking(test_digital_ocean_bucket())
+        put_head_get_list_delete_object_blocking(test_digital_ocean_bucket())
     }
 
     #[ignore]
